import datetime

_DEFAULT_CASH_VALUE = 1e8

class TradeTrigger:

    IndexRebalance = 1
    FundSelection = 10
    FundRebalance = 100
    FundEnd = 1000

    @staticmethod
    def _is(trigger, num):
        return trigger % (num * 10) / num == 1

    @staticmethod
    def is_end(trigger):
        return trigger / TradeTrigger.FundEnd >= 1

    @staticmethod
    def parse(trigger):
        _s = ''
        _s += 'Fe' if TradeTrigger.is_end(trigger) else '' # Fund End
        _s += 'Ir' if TradeTrigger._is(trigger, TradeTrigger.IndexRebalance) else '' # Index Rebalance  
        _s += 'Fs' if TradeTrigger._is(trigger, TradeTrigger.FundSelection) else '' # Fund Selection
        _s += 'Fr' if TradeTrigger._is(trigger, TradeTrigger.FundRebalance) else '' # Fund Rebalance
        if len(_s) == 0:
            _s = 'None'
        return _s

    @staticmethod
    def trigger_log(trigger:int, 
                    # Ir params
                    index_rebalance_max_diff:float,
                    _index_rebalance_max_diff_name:float,
                    JudgeIndexDiff:float,
                    #Fs params
                    fund_selection_min_score:float,
                    _fund_selection_min_name:str,
                    JudgeFundSelection:float,
                    fund_rank_list:list,
                    #Fr params
                    fund_rebalance_min_score:float,
                    _fund_rebalance_min_name:str,
                    JudgeFundRebalance:float,
              ):
        trigger_result = []
        if TradeTrigger.is_end(trigger):
            dic = {
                'rebalanace_reason': 'Fe',
                'index_id': _fund_rebalance_min_name,
                'rebalance_details': f'fund_rebalance_min_score: {round(fund_rebalance_min_score,3)} threshold: {JudgeFundRebalance} fund rank :{fund_rank_list}'
            }
            trigger_result.append(dic)
            return trigger_result

        if TradeTrigger._is(trigger, TradeTrigger.IndexRebalance):
            dic = {
                'rebalanace_reason': 'Ir',
                'index_id': _index_rebalance_max_diff_name,
                'rebalance_details': f'index_rebalance_max_diff: {round(index_rebalance_max_diff,3)} threshold: {JudgeIndexDiff}'
            }
            trigger_result.append(dic)
        
        if TradeTrigger._is(trigger, TradeTrigger.FundSelection):
            dic = {
                'rebalanace_reason': 'Fs',
                'index_id': _fund_selection_min_name,
                'rebalance_details': f'fund_selection_min_score: {round(fund_selection_min_score,3)} threshold: {JudgeFundSelection} fund rank :{fund_rank_list}'
            }
            trigger_result.append(dic)

        if TradeTrigger._is(trigger, TradeTrigger.FundRebalance):
            dic = {
                'rebalanace_reason': 'Fr',
                'index_id': _fund_rebalance_min_name,
                'rebalance_details': f'fund_rebalance_min_score: {round(fund_rebalance_min_score,3)} threshold: {JudgeFundRebalance}'
            }
            trigger_result.append(dic)
        
        return trigger_result

    @staticmethod
    def trigger_detail(trigger:int, 
                    # Ir params
                    index_rebalance_max_diff:float,
                    _index_rebalance_max_diff_name:float,
                    JudgeIndexDiff:float,
                    #Fs params
                    fund_selection_min_score:float,
                    _fund_selection_min_name:str,
                    JudgeFundSelection:float,
                    fund_rank_list:list,
                    #Fr params
                    fund_rebalance_min_score:float,
                    _fund_rebalance_min_name:str,
                    JudgeFundRebalance:float,
              ):
        result = {
            'index_rebalance_detail':   f'index_id {_index_rebalance_max_diff_name} index_rebalance_max_diff: {round(index_rebalance_max_diff,3)} threshold: {JudgeIndexDiff}',
            'fund_selection':           f'index_id {_fund_selection_min_name} fund_selection_min_score: {round(fund_selection_min_score,3)} threshold: {JudgeFundSelection} fund rank :{fund_rank_list}',
            'fund_rebalance':           f'index_id {_fund_rebalance_min_name} fund_rebalance_min_score {round(fund_rebalance_min_score,5)} threshold: {JudgeFundRebalance}'

        }
        return result



STOCK_BALANCE_MAPPING = {
    'DATES': '财报日期',
    # 流动资产部分
    'BALANCESTATEMENT_9': '货币资金',
    'BALANCESTATEMENT_67': '结算备付金',
    'BALANCESTATEMENT_181': '融出资金',
    'BALANCESTATEMENT_50': '拆出资金',
    'BALANCESTATEMENT_224': '交易性金融资产',
    'BALANCESTATEMENT_163': '以公允价值计量且其变动计入当期损益的金融资产',
    'BALANCESTATEMENT_10': '其中:交易性金融资产',
    'BALANCESTATEMENT_164': '其中:指定以公允价值计量且其变动计入当期损益的金融资产',
    'BALANCESTATEMENT_51': '衍生金融资产',
    'BALANCESTATEMENT_216': '应收票据及应收账款',
    'BALANCESTATEMENT_11': '应收票据',
    'BALANCESTATEMENT_12': '应收账款',
    'BALANCESTATEMENT_223': '应收款项融资',
    'BALANCESTATEMENT_14': '预付款项',
    'BALANCESTATEMENT_55': '应收保费',
    'BALANCESTATEMENT_57': '应收分保账款',
    'BALANCESTATEMENT_152': '应收分保合同准备金',
    'BALANCESTATEMENT_222': '其他应收款合计',
    'BALANCESTATEMENT_16': '其中:应收利息',
    'BALANCESTATEMENT_15': '其中:应收股利',
    'BALANCESTATEMENT_13': '其中:其他应收款',
    'BALANCESTATEMENT_156': '应收出口退税',
    'BALANCESTATEMENT_157': '应收补贴款',
    'BALANCESTATEMENT_158': '内部应收款',
    'BALANCESTATEMENT_52': '买入返售金融资产',
    'BALANCESTATEMENT_206': '以摊余成本计量的金融资产',
    'BALANCESTATEMENT_17': '存货',
    'BALANCESTATEMENT_207': '以公允价值计量且其变动计入其他综合收益的金融资产',
    'BALANCESTATEMENT_209': '合同资产',
    'BALANCESTATEMENT_202': '划分为持有待售的资产',
    'BALANCESTATEMENT_20': '一年内到期的非流动资产',
    'BALANCESTATEMENT_190': '代理业务资产',
    'BALANCESTATEMENT_21': '其他流动资产',
    'BALANCESTATEMENT_22': '流动资产其他项目',
    'BALANCESTATEMENT_23': '流动资产平衡项目',
    'BALANCESTATEMENT_25': '流动资产合计',
    # 非流动资产部分
    'BALANCESTATEMENT_53': '发放贷款及垫款',
    'BALANCESTATEMENT_217': '债权投资',
    'BALANCESTATEMENT_218': '其他债权投资',
    'BALANCESTATEMENT_211': '以摊余成本计量的金融资产(非流动)',
    'BALANCESTATEMENT_212': '以公允价值计量且其变动计入其他综合收益的金融资产(非流动)',
    'BALANCESTATEMENT_26': '可供出售金融资产',
    'BALANCESTATEMENT_27': '持有至到期投资',
    'BALANCESTATEMENT_30': '长期应收款',
    'BALANCESTATEMENT_29': '长期股权投资',
    'BALANCESTATEMENT_28': '投资性房地产',
    'BALANCESTATEMENT_31': '固定资产',
    'BALANCESTATEMENT_33': '在建工程',
    'BALANCESTATEMENT_32': '工程物资',
    'BALANCESTATEMENT_219': '其他权益工具投资',
    'BALANCESTATEMENT_220': '其他非流动金融资产',
    'BALANCESTATEMENT_34': '清理',
    'BALANCESTATEMENT_35': '生产性生物资产',
    'BALANCESTATEMENT_36': '油气资产',
    'BALANCESTATEMENT_225': '使用权资产',
    'BALANCESTATEMENT_37': '无形资产',
    'BALANCESTATEMENT_44': '非流动资产平衡项目',
    'BALANCESTATEMENT_38': '开发支出',
    'BALANCESTATEMENT_39': '商誉',
    'BALANCESTATEMENT_40': '长期待摊费用',
    'BALANCESTATEMENT_41': '递延所得税资产',
    'BALANCESTATEMENT_42': '其他非流动资产',
    'BALANCESTATEMENT_43': '非流动资产其他项目',
    'BALANCESTATEMENT_46': '非流动资产合计',
    'BALANCESTATEMENT_71': '资产其他项目',
    'BALANCESTATEMENT_72': '资产平衡项目',
    'BALANCESTATEMENT_74': '资产总计',
    # 流动负债
    'BALANCESTATEMENT_75': '短期借款',
    'BALANCESTATEMENT_105': '向中央银行借款',
    'BALANCESTATEMENT_153': '吸收存款及同业存放',
    'BALANCESTATEMENT_106': '拆入资金',
    'BALANCESTATEMENT_226': '交易性金融负债',
    'BALANCESTATEMENT_170': '以公允价值计量且其变动计入当期损益的金融负债',
    'BALANCESTATEMENT_76': '其中:交易性金融负债',
    'BALANCESTATEMENT_171': '其中:指定以公允价值计量且其变动计入当期损益的金融负债',
    'BALANCESTATEMENT_107': '衍生金融负债',
    'BALANCESTATEMENT_221': '应付票据及应付账款',
    'BALANCESTATEMENT_77': '应付票据',
    'BALANCESTATEMENT_78': '应付账款',
    'BALANCESTATEMENT_79': '预收款项',
    'BALANCESTATEMENT_213': '合同负债',
    'BALANCESTATEMENT_108': '卖出回购金融资产款',
    'BALANCESTATEMENT_113': '应付手续费及佣金',
    'BALANCESTATEMENT_80': '应付职工薪酬',
    'BALANCESTATEMENT_81': '应交税费',
    'BALANCESTATEMENT_227': '其他应付款合计',
    'BALANCESTATEMENT_82': '其中:应付利息',
    'BALANCESTATEMENT_83': '其中:应付股利',
    'BALANCESTATEMENT_84': '其中:其他应付款',
    'BALANCESTATEMENT_114': '应付分保账款',
    'BALANCESTATEMENT_188': '内部应付款',
    'BALANCESTATEMENT_189': '预计流动负债',
    'BALANCESTATEMENT_154': '保险合同准备金',
    'BALANCESTATEMENT_123': '代理买卖证券款',
    'BALANCESTATEMENT_124': '代理承销证券款',
    'BALANCESTATEMENT_87': '一年内的递延收益',
    'BALANCESTATEMENT_208': '以摊余成本计量的金融负债',
    'BALANCESTATEMENT_147': '应付短期债券',
    'BALANCESTATEMENT_203': '划分为持有待售的负债',
    'BALANCESTATEMENT_88': '一年内到期的非流动负债',
    'BALANCESTATEMENT_191': '代理业务负债',
    'BALANCESTATEMENT_89': '其他流动负债',
    'BALANCESTATEMENT_90': '流动负债其他项目',
    'BALANCESTATEMENT_91': '流动负债平衡项目',
    'BALANCESTATEMENT_93': '流动负债合计',
    # 非流动负债
    'BALANCESTATEMENT_94': '长期借款',
    'BALANCESTATEMENT_215': '以摊余成本计量的金融负债(非流动)',
    'BALANCESTATEMENT_95': '应付债券',
    'BALANCESTATEMENT_193': '其中:优先股(应付债券)',
    'BALANCESTATEMENT_194': '其中:永续债(应付债券)',
    'BALANCESTATEMENT_228': '租赁负债',
    'BALANCESTATEMENT_96': '长期应付款',
    'BALANCESTATEMENT_201': '长期应付职工薪酬',
    'BALANCESTATEMENT_97': '专项应付款',
    'BALANCESTATEMENT_86': '预计非流动负债',
    'BALANCESTATEMENT_148': '递延收益',
    'BALANCESTATEMENT_98': '递延所得税负债',
    'BALANCESTATEMENT_99': '其他非流动负债',
    'BALANCESTATEMENT_100': '非流动负债其他项目',
    'BALANCESTATEMENT_101': '非流动负债平衡项目',
    'BALANCESTATEMENT_103': '非流动负债合计',
    'BALANCESTATEMENT_125': '负债其他项目',
    'BALANCESTATEMENT_126': '负债平衡项目',
    'BALANCESTATEMENT_128': '负债合计',
    # 所有者权益
    'BALANCESTATEMENT_129': '实收资本(股本)',
    'BALANCESTATEMENT_195': '其他权益工具',
    'BALANCESTATEMENT_196': '其中:优先股(其他权益工具)',
    'BALANCESTATEMENT_197': '其中:永续债(其他权益工具)',
    'BALANCESTATEMENT_198': '其中:其他(其他权益工具)',
    'BALANCESTATEMENT_130': '资本公积',
    'BALANCESTATEMENT_199': '其他综合收益',
    'BALANCESTATEMENT_133': '库存股',
    'BALANCESTATEMENT_159': '专项储备',
    'BALANCESTATEMENT_131': '盈余公积',
    'BALANCESTATEMENT_134': '一般风险准备',
    'BALANCESTATEMENT_151': '未确定的投资损失',
    'BALANCESTATEMENT_132': '未分配利润',
    'BALANCESTATEMENT_160': '拟分配现金股利',
    'BALANCESTATEMENT_135': '外币报表折算差额',
    'BALANCESTATEMENT_161': '归属于母公司股东权益其他项目',
    'BALANCESTATEMENT_162': '归属于母公司股东权益平衡项目',
    'BALANCESTATEMENT_140': '归属于母公司股东权益合计',
    'BALANCESTATEMENT_136': '少数股东权益',
    'BALANCESTATEMENT_137': '股东权益其他项目',
    'BALANCESTATEMENT_138': '股东权益平衡项目',
    'BALANCESTATEMENT_141': '股东权益合计',
    'BALANCESTATEMENT_142': '负债和股东权益其他项目',
    'BALANCESTATEMENT_143': '负债和股东权益平衡项目',
    'BALANCESTATEMENT_145': '负债和股东权益合计',
}


STOCK_INCOME_MAPPING = {
    'DATES': '财报日期',
    'INCOMESTATEMENT_83': '营业总收入',
    'INCOMESTATEMENT_9': '营业收入',
    'INCOMESTATEMENT_19': '利息收入',
    'INCOMESTATEMENT_28': '已赚保费',
    'INCOMESTATEMENT_22': '手续费及佣金收入',
    'INCOMESTATEMENT_85': '其他业务收入',
    'INCOMESTATEMENT_88': '营业总收入其他项目',
    'INCOMESTATEMENT_84': '营业总成本',
    'INCOMESTATEMENT_10': '营业成本',
    'INCOMESTATEMENT_20': '利息支出',
    'INCOMESTATEMENT_23': '手续费及佣金支出',
    'INCOMESTATEMENT_89': '研发费用',
    'INCOMESTATEMENT_39': '退保金',
    'INCOMESTATEMENT_33': '赔付支出净额',
    'INCOMESTATEMENT_35': '提取保险合同准备金净额',
    'INCOMESTATEMENT_40': '保单红利支出',
    'INCOMESTATEMENT_38': '分保费用',
    'INCOMESTATEMENT_86': '其他业务成本',
    'INCOMESTATEMENT_11': '税金及附加',
    'INCOMESTATEMENT_12': '销售费用',
    'INCOMESTATEMENT_13': '管理费用',
    'INCOMESTATEMENT_14': '财务费用',
    'INCOMESTATEMENT_127': '其中:利息支出',
    'INCOMESTATEMENT_128': '其中:利息收入',
    'INCOMESTATEMENT_15': '资产减值损失',
    'INCOMESTATEMENT_129': '信用减值损失',
    'INCOMESTATEMENT_90': '营业总成本其他项目',
    'INCOMESTATEMENT_16': '公允价值变动收益',
    'INCOMESTATEMENT_17': '投资收益',
    'INCOMESTATEMENT_82': '对联营企业和合营企业的投资收益',
    'INCOMESTATEMENT_130': '净敞口套期收益',
    'INCOMESTATEMENT_25': '汇兑收益',
    'INCOMESTATEMENT_180': '资产减值损失(新)',
    'INCOMESTATEMENT_182': '信用减值损失(新)',
    'INCOMESTATEMENT_123': '资产处置收益',
    'INCOMESTATEMENT_124': '其他收益',
    'INCOMESTATEMENT_45': '营业利润其他项目',
    'INCOMESTATEMENT_46': '营业利润平衡项目',
    'INCOMESTATEMENT_48': '营业利润',
    'INCOMESTATEMENT_49': '营业外收入',
    'INCOMESTATEMENT_118': '非流动资产处置利得',
    'INCOMESTATEMENT_50': '营业外支出',
    'INCOMESTATEMENT_51': '非流动资产处置净损失',
    'INCOMESTATEMENT_52': '影响利润总额的其他项目',
    'INCOMESTATEMENT_53': '利润总额平衡项目',
    'INCOMESTATEMENT_55': '利润总额',
    'INCOMESTATEMENT_56': '所得税',
    'INCOMESTATEMENT_87': '未确认投资损失',
    'INCOMESTATEMENT_57': '影响净利润的其他项目',
    'INCOMESTATEMENT_120': '净利润差额(合计平衡项目2)',
    'INCOMESTATEMENT_60': '净利润',
    'INCOMESTATEMENT_125': '持续经营净利润',
    'INCOMESTATEMENT_126': '终止经营净利润',
    'INCOMESTATEMENT_91': '被合并方在合并前实现利润',
    'INCOMESTATEMENT_61': '归属于母公司股东的净利润',
    'INCOMESTATEMENT_62': '少数股东损益',
    'INCOMESTATEMENT_92': '净利润其他项目',
    'INCOMESTATEMENT_58': '净利润差额(合计平衡项目)',
    'INCOMESTATEMENT_80': '基本每股收益',
    'INCOMESTATEMENT_81': '稀释每股收益',
    'INCOMESTATEMENT_114': '其他综合收益',
    'INCOMESTATEMENT_116': '归属于母公司股东的其他综合收益',
    'INCOMESTATEMENT_115': '归属于少数股东的其他综合收益',
    'INCOMESTATEMENT_113': '综合收益总额',
    'INCOMESTATEMENT_93': '归属于母公司所有者的综合收益总额',
    'INCOMESTATEMENT_94': '归属于少数股东的综合收益总额',
    'INCOMESTATEMENT_139': '以摊余成本计量的金融资产终止确认收益',
}


STOCK_CASH_FLOW_MAPPING = {
    'DATES': '财报日期',
    'CASHFLOWSTATEMENT_9': '销售商品、提供劳务收到的现金',
    'CASHFLOWSTATEMENT_12': '客户存款和同业存放款项净增加额',
    'CASHFLOWSTATEMENT_13': '向中央银行借款净增加额',
    'CASHFLOWSTATEMENT_14': '向其他金融机构拆入资金净增加额',
    'CASHFLOWSTATEMENT_16': '收到原保险合同保费取得的现金',
    'CASHFLOWSTATEMENT_17': '收到再保险业务现金净额',
    'CASHFLOWSTATEMENT_64': '保户储金及投资款净增加额',
    'CASHFLOWSTATEMENT_18': '处置交易性金融资产净增加额',
    'CASHFLOWSTATEMENT_186': '收取利息、手续费及佣金的现金',
    'CASHFLOWSTATEMENT_20': '拆入资金净增加额',
    'CASHFLOWSTATEMENT_120': '发放贷款及垫款的净减少额',
    'CASHFLOWSTATEMENT_21': '回购业务资金净增加额',
    'CASHFLOWSTATEMENT_10': '收到的税费返还',
    'CASHFLOWSTATEMENT_11': '收到其他与经营活动有关的现金',
    'CASHFLOWSTATEMENT_22': '经营活动现金流入其他项目',
    'CASHFLOWSTATEMENT_23': '经营活动现金流入平衡项目',
    'CASHFLOWSTATEMENT_25': '经营活动现金流入小计',
    'CASHFLOWSTATEMENT_26': '购买商品、接受劳务支付的现金',
    'CASHFLOWSTATEMENT_30': '客户贷款及垫款净增加额',
    'CASHFLOWSTATEMENT_31': '存放中央银行和同业款项净增加额',
    'CASHFLOWSTATEMENT_32': '支付原保险合同赔付款项的现金',
    'CASHFLOWSTATEMENT_33': '支付利息、手续费及佣金的现金',
    'CASHFLOWSTATEMENT_195': '支付保单红利的现金',
    'CASHFLOWSTATEMENT_27': '支付给职工以及为职工支付的现金',
    'CASHFLOWSTATEMENT_28': '支付的各项税费',
    'CASHFLOWSTATEMENT_29': '支付其他与经营活动有关的现金',
    'CASHFLOWSTATEMENT_34': '经营活动现金流出其他项目',
    'CASHFLOWSTATEMENT_35': '经营活动现金流出平衡项目',
    'CASHFLOWSTATEMENT_37': '经营活动现金流出小计',
    'CASHFLOWSTATEMENT_121': '经营活动产生的现金流量净额其他项目',
    'CASHFLOWSTATEMENT_38': '经营活动产生的现金流量净额平衡项目',
    'CASHFLOWSTATEMENT_39': '经营活动产生的现金流量净额',
    # 投资活动产生的现金流量部分
    'CASHFLOWSTATEMENT_40': '收回投资收到的现金',
    'CASHFLOWSTATEMENT_41': '取得投资收益收到的现金',
    'CASHFLOWSTATEMENT_42': '处置固定资产、无形资产和其他长期资产收回的现金净额',
    'CASHFLOWSTATEMENT_43': '处置子公司及其他营业单位收到的现金净额',
    'CASHFLOWSTATEMENT_232': '减少质押和定期存款所收到的现金',
    'CASHFLOWSTATEMENT_44': '收到其他与投资活动有关的现金',
    'CASHFLOWSTATEMENT_45': '投资活动现金流入其他项目',
    'CASHFLOWSTATEMENT_46': '投资活动现金流入平衡项目',
    'CASHFLOWSTATEMENT_48': '投资活动现金流入小计',
    'CASHFLOWSTATEMENT_49': '购建固定资产、无形资产和其他长期资产支付的现金',
    'CASHFLOWSTATEMENT_50': '投资支付的现金',
    'CASHFLOWSTATEMENT_53': '质押贷款净增加额',
    'CASHFLOWSTATEMENT_51': '取得子公司及其他营业单位支付的现金净额',
    'CASHFLOWSTATEMENT_233': '增加质押和定期存款所支付的现金',
    'CASHFLOWSTATEMENT_52': '支付其他与投资活动有关的现金',
    'CASHFLOWSTATEMENT_54': '投资活动现金流出其他项目',
    'CASHFLOWSTATEMENT_55': '投资活动现金流出平衡项目',
    'CASHFLOWSTATEMENT_57': '投资活动现金流出小计',
    'CASHFLOWSTATEMENT_122': '投资活动产生的现金流量净额其他项目',
    'CASHFLOWSTATEMENT_58': '投资活动产生的现金流量净额平衡项目',
    'CASHFLOWSTATEMENT_59': '投资活动产生的现金流量净额',
    # 筹资活动产生的现金流量部分
    'CASHFLOWSTATEMENT_60': '吸收投资收到的现金',
    'CASHFLOWSTATEMENT_118': '其中:子公司吸收少数股东投资收到的现金',
    'CASHFLOWSTATEMENT_61': '取得借款收到的现金',
    'CASHFLOWSTATEMENT_63': '发行债券收到的现金',
    'CASHFLOWSTATEMENT_62': '收到其他与筹资活动有关的现金',
    'CASHFLOWSTATEMENT_65': '筹资活动现金流入其他项目',
    'CASHFLOWSTATEMENT_66': '筹资活动现金流入平衡项目',
    'CASHFLOWSTATEMENT_68': '筹资活动现金流入小计',
    'CASHFLOWSTATEMENT_69': '偿还债务支付的现金',
    'CASHFLOWSTATEMENT_70': '分配股利、利润或偿付利息支付的现金',
    'CASHFLOWSTATEMENT_119': '其中:子公司支付给少数股东的股利、利润',
    'CASHFLOWSTATEMENT_123': '购买子公司少数股权而支付的现金',
    'CASHFLOWSTATEMENT_71': '支付其他与筹资活动有关的现金',
    'CASHFLOWSTATEMENT_124': '其中:子公司减资支付给少数股东的现金',
    'CASHFLOWSTATEMENT_72': '筹资活动现金流出其他项目',
    'CASHFLOWSTATEMENT_73': '筹资活动现金流出平衡项目',
    'CASHFLOWSTATEMENT_75': '筹资活动现金流出小计',
    'CASHFLOWSTATEMENT_125': '筹资活动产生的现金流量净额其他项目',
    'CASHFLOWSTATEMENT_76': '筹资活动产生的现金流量净额平衡项目',
    'CASHFLOWSTATEMENT_77': '筹资活动产生的现金流量净额',
    # 现金及现金等价物净增加
    'CASHFLOWSTATEMENT_78': '汇率变动对现金及现金等价物的影响',
    'CASHFLOWSTATEMENT_79': '现金及现金等价物净增加额其他项目',
    'CASHFLOWSTATEMENT_80': '现金及现金等价物净增加额平衡项目',
    'CASHFLOWSTATEMENT_82': '现金及现金等价物净增加额',
    'CASHFLOWSTATEMENT_83': '期初现金及现金等价物余额',
    'CASHFLOWSTATEMENT_126': '期末现金及现金等价物余额其他项目',
    'CASHFLOWSTATEMENT_127': '期末现金及现金等价物余额平衡项目',
    'CASHFLOWSTATEMENT_84': '期末现金及现金等价物余额',
    # 补充资料
    'CASHFLOWSTATEMENT_85': '净利润-现金流量表',
    'CASHFLOWSTATEMENT_86': '资产减值准备',
    'CASHFLOWSTATEMENT_206': '固定资产和投资性房地产折旧',
    'CASHFLOWSTATEMENT_87': '其中：固定资产折旧、油气资产折耗、生产性生物资产折旧',
    'CASHFLOWSTATEMENT_207': '其中：投资性房地产折旧',
    'CASHFLOWSTATEMENT_88': '无形资产摊销',
    'CASHFLOWSTATEMENT_89': '长期待摊费用摊销',
    'CASHFLOWSTATEMENT_208': '递延收益摊销',
    'CASHFLOWSTATEMENT_90': '待摊费用的减少',
    'CASHFLOWSTATEMENT_91': '预提费用的增加',
    'CASHFLOWSTATEMENT_92': '处置固定资产、无形资产和其他长期资产的损失',
    'CASHFLOWSTATEMENT_93': '固定资产报废损失',
    'CASHFLOWSTATEMENT_94': '公允价值变动损失',
    'CASHFLOWSTATEMENT_95': '财务费用-现金流量表',
    'CASHFLOWSTATEMENT_96': '投资损失',
    'CASHFLOWSTATEMENT_209': '递延所得税',
    'CASHFLOWSTATEMENT_97': '其中：递延所得税资产减少',
    'CASHFLOWSTATEMENT_98': '其中：递延所得税负债增加',
    'CASHFLOWSTATEMENT_210': '预计负债的增加',
    'CASHFLOWSTATEMENT_99': '存货的减少',
    'CASHFLOWSTATEMENT_100': '经营性应收项目的减少',
    'CASHFLOWSTATEMENT_101': '经营性应付项目的增加',
    'CASHFLOWSTATEMENT_117': '其他',
    'CASHFLOWSTATEMENT_102': '经营活动产生的现金流量净额其他项目',
    'CASHFLOWSTATEMENT_103': '经营活动产生的现金流量净额平衡项目',
    'CASHFLOWSTATEMENT_105': '间接法—经营活动产生的现金流量净额',
    'CASHFLOWSTATEMENT_106': '债务转为资本',
    'CASHFLOWSTATEMENT_107': '一年内到期的可转换公司债券',
    'CASHFLOWSTATEMENT_108': '融资租入固定资产',
    'CASHFLOWSTATEMENT_212': '不涉及现金收支的投资和筹资活动金额其他项目',
    'CASHFLOWSTATEMENT_109': '现金的期末余额',
    'CASHFLOWSTATEMENT_110': '现金的期初余额',
    'CASHFLOWSTATEMENT_111': '现金等价物的期末余额',
    'CASHFLOWSTATEMENT_112': '现金等价物的期初余额',
    'CASHFLOWSTATEMENT_113': '现金及现金等价物净增加额其他项目',
    'CASHFLOWSTATEMENT_114': '现金及现金等价物净增加额平衡项目',
    'CASHFLOWSTATEMENT_116': '间接法—现金及现金等价物的净增加额',
}


STOCK_INCOME_QUARTERLY_MAPPING = {
    'DATES': '财报日期',
    'INCOMESTATEMENTQ_83': '单季度.营业总收入',
    'INCOMESTATEMENTQ_9': '单季度.营业收入',
    'INCOMESTATEMENTQ_19': '单季度.利息收入',
    'INCOMESTATEMENTQ_28': '单季度.已赚保费',
    'INCOMESTATEMENTQ_22': '单季度.手续费及佣金收入',
    'INCOMESTATEMENTQ_85': '单季度.其他业务收入',
    'INCOMESTATEMENTQ_88': '单季度.营业总收入其他项目',
    'INCOMESTATEMENTQ_84': '单季度.营业总成本',
    'INCOMESTATEMENTQ_10': '单季度.营业成本',
    'INCOMESTATEMENTQ_20': '单季度.利息支出',
    'INCOMESTATEMENTQ_23': '单季度.手续费及佣金支出',
    'INCOMESTATEMENTQ_89': '单季度.研发费用',
    'INCOMESTATEMENTQ_39': '单季度.退保金',
    'INCOMESTATEMENTQ_33': '单季度.赔付支出净额',
    'INCOMESTATEMENTQ_35': '单季度.提取保险合同准备金净额',
    'INCOMESTATEMENTQ_40': '单季度.保单红利支出',
    'INCOMESTATEMENTQ_38': '单季度.分保费用',
    'INCOMESTATEMENTQ_86': '单季度.其他业务成本',
    'INCOMESTATEMENTQ_11': '单季度.税金及附加',
    'INCOMESTATEMENTQ_12': '单季度.销售费用',
    'INCOMESTATEMENTQ_13': '单季度.管理费用',
    'INCOMESTATEMENTQ_14': '单季度.财务费用',
    'INCOMESTATEMENTQ_127': '单季度.其中:利息费用',
    'INCOMESTATEMENTQ_128': '单季度.其中:利息收入',
    'INCOMESTATEMENTQ_15': '单季度.资产减值损失',
    'INCOMESTATEMENTQ_129': '单季度.信用减值损失',
    'INCOMESTATEMENTQ_90': '单季度.营业总成本其他项目',
    'INCOMESTATEMENTQ_16': '单季度.公允价值变动净收益',
    'INCOMESTATEMENTQ_17': '单季度.投资净收益',
    'INCOMESTATEMENTQ_82': '单季度.对联营企业和合营企业的投资收益',
    'INCOMESTATEMENTQ_130': '单季度.净敞口套期收益',
    'INCOMESTATEMENTQ_25': '单季度.汇兑收益',
    'INCOMESTATEMENTQ_123': '单季度.资产处置收益',
    'INCOMESTATEMENTQ_124': '单季度.其他收益',
    'INCOMESTATEMENTQ_45': '单季度.营业利润其他项目',
    'INCOMESTATEMENTQ_46': '单季度.营业利润平衡项目',
    'INCOMESTATEMENTQ_48': '单季度.营业利润',
    'INCOMESTATEMENTQ_49': '单季度.营业外收入',
    'INCOMESTATEMENTQ_118': '单季度.非流动资产处置利得',
    'INCOMESTATEMENTQ_50': '单季度.营业外支出',
    'INCOMESTATEMENTQ_51': '单季度.非流动资产处置净损失',
    'INCOMESTATEMENTQ_52': '单季度.影响利润总额的其他项目',
    'INCOMESTATEMENTQ_53': '单季度.利润总额平衡项目',
    'INCOMESTATEMENTQ_55': '单季度.利润总额',
    'INCOMESTATEMENTQ_56': '单季度.所得税费用',
    'INCOMESTATEMENTQ_87': '单季度.未确认的投资损失',
    'INCOMESTATEMENTQ_57': '单季度.影响净利润的其他项目',
    'INCOMESTATEMENTQ_120': '单季度.净利润差额(合计平衡项目2)',
    'INCOMESTATEMENTQ_60': '单季度.净利润',
    'INCOMESTATEMENTQ_125': '单季度.持续经营净利润',
    'INCOMESTATEMENTQ_126': '单季度.终止经营净利润',
    'INCOMESTATEMENTQ_91': '单季度.被合并方在合并前实现利润',
    'INCOMESTATEMENTQ_61': '单季度.归属于母公司股东的净利润',
    'INCOMESTATEMENTQ_62': '单季度.少数股东损益',
    'INCOMESTATEMENTQ_92': '单季度.净利润其他项目',
    'INCOMESTATEMENTQ_80': '单季度.基本每股收益',
    'INCOMESTATEMENTQ_81': '单季度.稀释每股收益',
    'INCOMESTATEMENTQ_114': '单季度.其他综合收益',
    'INCOMESTATEMENTQ_116': '单季度.归属于母公司股东的其他综合收益',
    'INCOMESTATEMENTQ_115': '单季度.归属于少数股东的其他综合收益总额',
    'INCOMESTATEMENTQ_113': '单季度.综合收益总额',
    'INCOMESTATEMENTQ_93': '单季度.归属于母公司所有者的综合收益总额',
    'INCOMESTATEMENTQ_94': '单季度.归属于少数股东的综合收益总额',
    'INCOMESTATEMENTQ_182': '单季度.信用减值损失(新)',
    'INCOMESTATEMENTQ_180': '单季度.资产减值损失(新)',
}


STOCK_CASH_FLOW_QUARTERLY_MAPPING = {
    'DATES': '财报日期',
    'CASHFLOWSTATEMENTQ_9': '单季度.销售商品、提供劳务收到的现金',
    'CASHFLOWSTATEMENTQ_12': '单季度.客户存款和同业存放款项净增加额',
    'CASHFLOWSTATEMENTQ_13': '单季度.向中央银行借款净增加额',
    'CASHFLOWSTATEMENTQ_14': '单季度.向其他金融机构拆入资金净增加额',
    'CASHFLOWSTATEMENTQ_16': '单季度.收到原保险合同保费取得的现金',
    'CASHFLOWSTATEMENTQ_17': '单季度.收到再保险业务现金净额',
    'CASHFLOWSTATEMENTQ_64': '单季度.保户储金及投资款净增加额',
    'CASHFLOWSTATEMENTQ_18': '单季度.处置交易性金融资产净增加额',
    'CASHFLOWSTATEMENTQ_186': '单季度.收取利息、手续费及佣金的现金',
    'CASHFLOWSTATEMENTQ_20': '单季度.拆入资金净增加额',
    'CASHFLOWSTATEMENTQ_120': '单季度.发放贷款及垫款的净减少额',
    'CASHFLOWSTATEMENTQ_21': '单季度.回购业务资金净增加额',
    'CASHFLOWSTATEMENTQ_10': '单季度.收到的税费返还',
    'CASHFLOWSTATEMENTQ_11': '单季度.收到的其他与经营活动有关的现金',
    'CASHFLOWSTATEMENTQ_22': '单季度.经营活动现金流入其他项目',
    'CASHFLOWSTATEMENTQ_23': '单季度.经营活动现金流入平衡项目',
    'CASHFLOWSTATEMENTQ_25': '单季度.经营活动现金流入小计',
    'CASHFLOWSTATEMENTQ_26': '单季度.购买商品、接受劳务支付的现金',
    'CASHFLOWSTATEMENTQ_30': '单季度.客户贷款及垫款净增加额',
    'CASHFLOWSTATEMENTQ_31': '单季度.存放中央银行和同业款项净增加额',
    'CASHFLOWSTATEMENTQ_32': '单季度.支付原保险合同赔付款项的现金',
    'CASHFLOWSTATEMENTQ_33': '单季度.支付利息、手续费及佣金的现金',
    'CASHFLOWSTATEMENTQ_195': '单季度.支付保单红利的现金',
    'CASHFLOWSTATEMENTQ_27': '单季度.支付给职工以及为职工支付的现金',
    'CASHFLOWSTATEMENTQ_28': '单季度.支付的各项税费',
    'CASHFLOWSTATEMENTQ_29': '单季度.支付的其他与经营活动有关的现金',
    'CASHFLOWSTATEMENTQ_34': '单季度.经营活动现金流出其他项目',
    'CASHFLOWSTATEMENTQ_35': '单季度.经营活动现金流出平衡项目',
    'CASHFLOWSTATEMENTQ_37': '单季度.经营活动现金流出小计',
    'CASHFLOWSTATEMENTQ_121': '单季度.经营活动产生的现金流量净额其他项目',
    'CASHFLOWSTATEMENTQ_38': '单季度.经营活动产生的现金流量净额平衡项目',
    'CASHFLOWSTATEMENTQ_39': '单季度.经营活动产生的现金流量净额',
    'CASHFLOWSTATEMENTQ_192': '银行业务及证券业务拆借资金净减少额',
    # 投资活动产生的现金流量
    'CASHFLOWSTATEMENTQ_40': '单季度.收回投资收到的现金',
    'CASHFLOWSTATEMENTQ_41': '单季度.取得投资收益收到的现金',
    'CASHFLOWSTATEMENTQ_42': '单季度.处置固定资产、无形资产和其他长期资产所收回的现金净额',
    'CASHFLOWSTATEMENTQ_43': '单季度.处置子公司及其他营业单位收到的现金净额',
    'CASHFLOWSTATEMENTQ_232': '单季度.减少质押和定期存款所收到的现金',
    'CASHFLOWSTATEMENTQ_44': '单季度.收到的其他与投资活动有关的现金',
    'CASHFLOWSTATEMENTQ_45': '单季度.投资活动现金流入其他项目',
    'CASHFLOWSTATEMENTQ_46': '单季度.投资活动现金流入平衡项目',
    'CASHFLOWSTATEMENTQ_48': '单季度.投资活动现金流入小计',
    'CASHFLOWSTATEMENTQ_49': '单季度.购建固定资产、无形资产和其他长期资产所支付的现金',
    'CASHFLOWSTATEMENTQ_50': '单季度.投资所支付的现金',
    'CASHFLOWSTATEMENTQ_53': '单季度.质押贷款净增加额',
    'CASHFLOWSTATEMENTQ_51': '单季度.取得子公司及其他营业单位支付的现金净额',
    'CASHFLOWSTATEMENTQ_233': '单季度.增加质押和定期存款所支付的现金',
    'CASHFLOWSTATEMENTQ_52': '单季度.支付的其他与投资活动有关的现金',
    'CASHFLOWSTATEMENTQ_54': '单季度.投资活动现金流出其他项目',
    'CASHFLOWSTATEMENTQ_57': '单季度.投资活动现金流出小计',
    'CASHFLOWSTATEMENTQ_55': '单季度.投资活动现金流出平衡项目',
    'CASHFLOWSTATEMENTQ_122': '单季度.投资活动产生的现金流量净额其他项目',
    'CASHFLOWSTATEMENTQ_59': '单季度.投资活动产生的现金流量净额',
    'CASHFLOWSTATEMENTQ_58': '单季度.投资活动产生的现金流量净额平衡项目',
    # 筹资活动产生的现金流量
    'CASHFLOWSTATEMENTQ_60': '单季度.吸收投资收到的现金',
    'CASHFLOWSTATEMENTQ_118': '单季度.子公司吸收少数股东投资收到的现金',
    'CASHFLOWSTATEMENTQ_61': '单季度.取得借款收到的现金',
    'CASHFLOWSTATEMENTQ_63': '单季度.发行债券收到的现金',
    'CASHFLOWSTATEMENTQ_62': '单季度.收到其他与筹资活动有关的现金',
    'CASHFLOWSTATEMENTQ_65': '单季度.筹资活动现金流入其他项目',
    'CASHFLOWSTATEMENTQ_66': '单季度.筹资活动现金流入平衡项目',
    'CASHFLOWSTATEMENTQ_68': '单季度.筹资活动现金流入小计',
    'CASHFLOWSTATEMENTQ_69': '单季度.偿还债务支付的现金',
    'CASHFLOWSTATEMENTQ_70': '单季度.分配股利、利润或偿付利息所支付的现金',
    'CASHFLOWSTATEMENTQ_119': '单季度.子公司支付给少数股东的股利、利润',
    'CASHFLOWSTATEMENTQ_123': '单季度.购买子公司少数股权而支付的现金',
    'CASHFLOWSTATEMENTQ_71': '单季度.支付其他与筹资活动有关的现金',
    'CASHFLOWSTATEMENTQ_124': '单季度.子公司减资支付给少数股东的现金',
    'CASHFLOWSTATEMENTQ_72': '单季度.筹资活动现金流出其他项目',
    'CASHFLOWSTATEMENTQ_73': '单季度.筹资活动现金流出平衡项目',
    'CASHFLOWSTATEMENTQ_75': '单季度.筹资活动现金流出小计',
    'CASHFLOWSTATEMENTQ_125': '单季度.筹资活动产生的现金流量净额其他项目',
    'CASHFLOWSTATEMENTQ_76': '单季度.筹资活动产生的现金流量净额平衡项目',
    'CASHFLOWSTATEMENTQ_77': '单季度.筹资活动产生的现金流量净额',
    # 现金及现金等价物净增加值
    'CASHFLOWSTATEMENTQ_78': '单季度.汇率变动对现金及现金等价物的影响',
    'CASHFLOWSTATEMENTQ_79': '单季度.现金及现金等价物净增加额其他项目',
    'CASHFLOWSTATEMENTQ_80': '单季度.现金及现金等价物净增加额平衡项目',
    'CASHFLOWSTATEMENTQ_82': '单季度.现金及现金等价物净增加额',
    'CASHFLOWSTATEMENTQ_83': '单季度.期初现金及现金等价物余额',
    'CASHFLOWSTATEMENTQ_126': '单季度.期末现金及现金等价物余额其他项目',
    'CASHFLOWSTATEMENTQ_127': '单季度.期末现金及现金等价物余额平衡项目',
    'CASHFLOWSTATEMENTQ_84': '单季度.期末现金及现金等价物余额',
}


USED_BALANCE_COLUMNS = [
    '货币资金'
]

ROE有效阈值 = 0.8







from typing import Optional, List, Dict
from types import FunctionType
import os
import traceback
import re
import io
import json

import pandas as pd
import numpy as np
import pdfplumber

from ...constant import EMailParserType
from ..api.basic import BasicDataApi


EMAIL_DATA_COLUMNS_DICT = {
    '基金代码': 'fof_id',
    '产品代码': 'fof_id',
    '资产代码': 'fof_id',
    '基金名称': 'fof_name',
    '产品名称': 'fof_name',
    '资产名称': 'fof_name',
    '账套名称': 'fof_name',
    '基金份额净值': 'nav',
    '单位净值': 'nav',
    '计提前单位净值': 'nav',
    '实际净值': 'nav',
    '基金单位净值': 'nav',
    '资产份额净值(元)': 'nav',
    '本级别单位净值': 'nav',
    '基金份额累计净值': 'acc_net_value',
    '累计单位净值': 'acc_net_value',
    '累计净值': 'acc_net_value',
    '资产份额累计净值(元)': 'acc_net_value',
    '本级别累计单位净值': 'acc_net_value',
    '虚拟后净值': 'v_nav',
    '虚拟净值': 'v_nav',
    '计提后单位净值': 'v_nav',
    '虚拟计提净值': 'v_nav',
    '虚拟后单位净值': 'v_nav',
    '日期': 'datetime',
    '净值日期': 'datetime',
    '业务日期': 'datetime',
    '估值日期': 'datetime',
}


class LogCollector:

    def __init__(self):
        # 暂时先简单就是个 list
        self._log_container: List[str] = []

    def log(self, content: str):
        print(content)
        self._log_container.append(content)

    def get_out(self) -> List[str]:
        return self._log_container


class ParserRunner(LogCollector):

    def __init__(self, manager_id: str, parser_type: EMailParserType):
        super().__init__()

        self._manager_id = manager_id
        self._parser_type = parser_type

        parsers_manager: Optional[pd.DataFrame] = BasicDataApi().get_email_parser_exec(manager_id)
        assert parsers_manager is not None and not parsers_manager.empty, f'get email parsers exec info of {manager_id} failed (parser_type){parser_type}'

        if parsers_manager.shape[0] > 1:
            parsers_manager = parsers_manager[parsers_manager.manager_id.notna()]

        priority_info = json.loads(parsers_manager.p_priority_info.array[0])
        assert str(parser_type) in priority_info, f'no {parser_type} exec info of {manager_id}'
        parsers_list_with_priority: List[str] = priority_info[str(parser_type)]
        print(f'(manager_id){manager_id} {parsers_list_with_priority}')

        parsers_code: Optional[pd.DataFrame] = BasicDataApi().get_email_parser_code(parsers_list_with_priority)
        assert parsers_code is not None and not parsers_code.empty, f'get email parsers code of {manager_id} failed (parser_type){parser_type}'

        self._parsers: Dict[str, str] = parsers_code.set_index('id')['p_code'].to_dict()
        self.log(f'got parsers {self._parsers.keys()} to run (parser_type){parser_type}')

    def parse_data(self, datas: io.BytesIO, file_name: str = '') -> Optional[pd.DataFrame]:
        for p_id, p_code in self._parsers.items():
            self.log(f'try to run parser {p_id}')
            try:
                # FIXME 这里这样写可能不太好
                the_parser_code = compile(p_code, '<string>', 'exec')
                if the_parser_code.co_consts[0] is None:
                    self.log(f'can not find func in parser {p_id} for {file_name}, try the next one')
                    continue
                the_parser = FunctionType(the_parser_code.co_consts[0], globals())
                if self._parser_type == EMailParserType.E_PARSER_NAV:
                    df = the_parser(datas, file_name)
                elif self._parser_type == EMailParserType.E_PARSER_VALUATION:
                    df = the_parser(datas, self._manager_id)
                else:
                    df = None
                self.log(df)
                return df
            except Exception:
                # traceback.print_exc()
                self.log(f'parse {file_name} failed with parser {p_id}, try the next one')
            else:
                self.log(f'done, parse {file_name} succeed, by parser {p_id}')
                break
        else:
            self.log(f'parse {file_name} failed finally')

    def parse_file(self, path, file_name: str) -> Optional[pd.DataFrame]:
        with open(path, 'rb') as datas:
            return self.parse_data(datas, file_name)


class ParserRunnerTest(ParserRunner):

    def __init__(self, manager_id: str, parser_type: EMailParserType, test_case_dir: str):
        super().__init__(manager_id, parser_type)

        self._test_case_dir: str = test_case_dir
        assert os.path.isdir(self._test_case_dir), f'arg test_case_dir should be a directory (now){self._test_case_dir}'

    def run(self):
        with os.scandir(self._test_case_dir) as it:
            for entry in it:
                if not entry.name.startswith('.') and entry.is_file():
                    self.parse_file(entry.path, entry.name)


if __name__ == '__main__':
    MANAGER_ID = 'DEMO001'

    prt = ParserRunnerTest(MANAGER_ID, EMailParserType.E_PARSER_NAV, './test_parser')
    # prt.run()



import pandas as pd
import numpy as np
import datetime
import traceback
import time
from typing import List, Dict, Optional
from functools import partial
from multiprocessing import Pool

from ....util.calculator import Calculator
from ...api.basic import BasicDataApi
from ...api.derived import DerivedDataApi
from ...wrapper.mysql import BasicDatabaseConnector, DerivedDatabaseConnector
from ...view.basic_models import FundNav, Fund_size_and_hold_rate, FundInfo, FundStatusLatest, FundOpenInfo
from ...view.derived_models import FundManagerInfo, FundManagerIndex, FundManagerScore
from .derived_data_helper import DerivedDataHelper, FUND_CLASSIFIER, normalize, score_rescale
from ...manager.fund_info_filter import EHR_PROBLEM_EKYWORDS
 
class ManagerProcessor:

    INDEX_LIST = {'csi_boodfund': 'H11023.CSI', 'csi_f_hybrid': 'H11022.CSI', 'csi_f_qdii': 'H11026.CSI', 'csi_stockfund': 'H11021.CSI', 'mmf': 'H11025.CSI'}
    REPLACE_DICT = {'stock': 'H11021.CSI', 'bond': 'H11023.CSI', 'mmf': 'H11025.CSI', 'QDII': 'H11026.CSI', 'mix': 'H11022.CSI'}
    DATE_DICT = {'stock': datetime.date(2007,1,1), 'bond': datetime.date(2003,1,1), 'mmf': datetime.date(2004,1,7), 'QDII': datetime.date(2008,1,1), 'mix': datetime.date(2007,1,1)}

    def __init__(self, data_helper: DerivedDataHelper):
        self._data_helper: DerivedDataHelper = data_helper
        self._basic_api = BasicDataApi()
        self.mng_index_list: Dict[str, pd.DataFrame] = {}  # 基金经理指数
        self.mng_indicator_list: Dict[str, pd.DataFrame] = {}  # 基金经理评分

    def init(self, start_date: str, end_date: datetime.date, print_time=False, is_mng_index=False, is_mng_indicator=False):
        self.start_date = start_date if isinstance(start_date, datetime.date) else pd.to_datetime(start_date, infer_datetime_format=True).date()
        self.end_date = end_date if isinstance(end_date, datetime.date) else pd.to_datetime(end_date, infer_datetime_format=True).date()

        with BasicDatabaseConnector().managed_session() as quant_session:
            # 交易日序列
            self.trading_days: pd.DataFrame = self._basic_api.get_trading_day_list()
            self.trading_days_list = self.trading_days.datetime
            self.trim_days = self.trading_days[(self.trading_days.datetime >= self.start_date) & (self.trading_days.datetime <= self.end_date)].datetime

            # 指数价格
            self.index_price = self._basic_api.get_index_price(index_list=list(self.INDEX_LIST.keys()))
            self.index_price['index_id'] = self.index_price['index_id'].map(self.INDEX_LIST)
            self.index_price = self.index_price.pivot(index='datetime', columns='index_id', values='close').reindex(self.trim_days).ffill().dropna(how='all', axis=1)

            # 场内基金列表
            fund_status_query = quant_session.query(FundStatusLatest)
            self.fund_status = pd.read_sql(fund_status_query.statement, fund_status_query.session.bind)
            self.in_market_fund_list = self.fund_status[~self.fund_status.trade_status.isnull()].fund_id.tolist()

            # 封闭基金列表
            fund_open_info_query = quant_session.query(FundOpenInfo)
            self.fund_close = pd.read_sql(fund_open_info_query.statement, fund_open_info_query.session.bind)
            self.close_fund_list = self.fund_close.fund_id.tolist()

            # 基金信息
            fund_info_query = quant_session.query(FundInfo)
            self.fund_info: pd.DataFrame = pd.read_sql(fund_info_query.statement, fund_info_query.session.bind)

            # EHR类基金
            self.ehr_funds = []
            for r in self.fund_info[['desc_name', 'fund_id']].itertuples():
                for type_word, problem_words in EHR_PROBLEM_EKYWORDS.items():
                    i = r.desc_name
                    for problem_i in problem_words:
                        i = i.replace(problem_i, '')
                    if type_word in i:
                        self.ehr_funds.append(r.fund_id)
                    
            ## 计算基金经理指数时只去掉分级基金子基金， 不能回测交易的基金依然影响基金经理能力
            self.fund_info = self.fund_info[self.fund_info.structure_type <= 1]
            # 基金规模 报告期数据填充交易日 然后再只取交易日，保证和基金收益序列一致
            fund_size_query = quant_session.query(
                Fund_size_and_hold_rate.fund_id,
                Fund_size_and_hold_rate.datetime,
                Fund_size_and_hold_rate.size
            )
            self.fund_size = pd.read_sql(fund_size_query.statement, fund_size_query.session.bind)
            self.fund_size = self.fund_size.pivot_table(index='datetime', columns='fund_id', values='size')
            self.fund_size = self.fund_size.reindex(self.fund_size.index.union(self.trim_days)).dropna(axis=1, how='all')
            self.fund_size = self.fund_size.ffill()# + self.fund_size.bfill() - self.fund_size.bfill()

            if is_mng_index:
                # 基金净值 和 日收益
                fund_nav_query = quant_session.query(
                    FundNav.fund_id,
                    FundNav.adjusted_net_value,
                    FundNav.datetime
                ).filter(
                    FundNav.datetime >= self.start_date,
                    FundNav.datetime <= self.end_date,
                    FundNav.fund_id.in_(self.fund_info.fund_id)
                )
                self.fund_nav = pd.read_sql(fund_nav_query.statement, fund_nav_query.session.bind)
                self.fund_nav = self.fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').reindex(self.trim_days).dropna(axis=1, how='all')
                self.fund_nav = self.fund_nav.ffill() + self.fund_nav.bfill() - self.fund_nav.bfill()
                self.fund_ret = self.fund_nav.pct_change(fill_method=None)
                self.fund_size = self.fund_size.reindex(self.fund_ret.index)

            elif is_mng_indicator:
                self.fund_manager_index: pd.DataFrame = DerivedDataApi().get_fund_manager_index().drop(columns='_update_time')
                self.fund_manager_index.groupby(by='fund_type', sort=False).apply(self._preprocess_manager_index_df)
                self.fund_size = self.fund_size.reindex(self.trim_days)

        with DerivedDatabaseConnector().managed_session() as quant_session:
            # 基金经理信息
            mng_query = quant_session.query(
                FundManagerInfo.mng_id,
                FundManagerInfo.start_date,
                FundManagerInfo.fund_id,
                FundManagerInfo.mng_name,
                FundManagerInfo.end_date)
            self.fund_manager_info: pd.DataFrame = pd.read_sql(mng_query.statement, mng_query.session.bind)
            self.fund_manager_info = self.fund_manager_info.rename(columns={'mng_id': 'manager_id', 'mng_name': 'name', 'start_date': 'start', 'end_date': 'end'})
            self.fund_manager_info = self.fund_manager_info[self.fund_manager_info.fund_id != 'not_exist']
            self.fund_manager_info = pd.merge(self.fund_manager_info, self.fund_info[['fund_id','company_id']], how='left', on='fund_id' )

        self.manager_name_dict = dict(zip(self.fund_manager_info['manager_id'].tolist(), self.fund_manager_info['name'].tolist()))
        # 基金经理数据 根据资产分类
        if is_mng_index:
            self.class_fund = {}
            self.class_index: Dict[str, pd.DataFrame] = {}
            self.class_info: Dict[str, pd.DataFrame] = {}
            self.manager_name_dict = dict(zip(self.fund_manager_info['manager_id'].tolist(), self.fund_manager_info['name'].tolist()))
            for class_name, index_id in self.REPLACE_DICT.items():
                self.mng_index_list[class_name] = pd.DataFrame(index=self.fund_ret.index)
                fund = self.fund_info[self.fund_info['wind_class_2'].isin(FUND_CLASSIFIER[class_name])].fund_id
                info = self.fund_manager_info.loc[self.fund_manager_info['fund_id'].isin(fund), :]
                info = info[info['fund_id'].isin(self.fund_ret.columns.intersection(self.fund_size.columns))]
                self.class_info[class_name] = info
                self.class_index[class_name] = self.index_price[index_id].pct_change(fill_method=None)
                self.class_fund[class_name] = fund

        if is_mng_indicator:
            self.class_fund = {}
            for class_name, index_id in self.REPLACE_DICT.items():
                self.class_fund[class_name] = self.fund_info[self.fund_info['wind_class_2'].isin(FUND_CLASSIFIER[class_name])].fund_id

    @staticmethod
    def _lambda_func_1(single_fund_info: pd.DataFrame, x: pd.Series) -> pd.Series:
        result: List[pd.Series] = []
        for row in single_fund_info.itertuples(index=False):
            result.append(pd.Series((x.index >= row.start) & (x.index <= row.end), index=x.index))
        return pd.concat(result, axis=1).sum(axis=1) > 0

    def calc_index(self):
        # 遍历每个基金经理
        #_tm_part2_start = time.time()
        for m_id in self.manager_name_dict:# 遍历每个资产类别
            for class_name in self.REPLACE_DICT:
                idx = self.index_calculation(m_id, class_name)
                if idx is not None:
                    self.mng_index_list[class_name].loc[:, m_id] = idx # 该资产类别下记录计算结果
        for class_name in self.REPLACE_DICT:
            self.mng_index_list[class_name] = self.mng_index_list[class_name].replace(0, np.nan).dropna(how='all')
        # _tm_part2_end = time.time()
        # print(f' part2 cost time {_tm_part2_end - _tm_part2_start}')

    def index_calculation(self, manager: str, class_name: str):
        '''
        计算基金经理指数
        manager：基金经理的ID
        fund_kind：基金的分类（'stock', 'bond', 'money', 'mix', 'QDII'）
        replace_index：空窗期选择替代的指数的代码
        rate_df：基金日收益率的dataframe
        scale_df：基金规模的dataframe
        返回基金经理指数的Series，时间是从第一个产品的任职到最后一个产品的离职
        replace_index为空时，空窗期按照_REPLACE_DICT中的指数来代替（如果没有指定基金的分类，按照沪深300的指数）
        基金规模按照上期季报的规模作为权重
        如果该日基金净值缺失，则用之前的净值代替，如果历史上并没有该日的净值数据，那么在计算基金经理指数的过程中该基金产品不再拥有权重
        如果因该日基金净值缺失而导致该日基金经理没有基金产品可供计算指数，按照空窗期来处理
        '''
        class_info = self.class_info[class_name]
        m_info = class_info[class_info['manager_id'] == manager]
        index_rate = self.class_index[class_name]
        if m_info.shape[0] == 0:
            # raise Exception('该经理没有相关基金产品')
            return
        m_start = m_info['start'].min()
        m_end = m_info['end'].max()
        unique_fund_ids = m_info['fund_id'].unique()
        rate: pd.DataFrame = self.fund_ret.loc[m_start:m_end, unique_fund_ids]
        scale: pd.DataFrame = self.fund_size.loc[m_start:m_end, unique_fund_ids]
        for fund_id in unique_fund_ids:
            single_fund_info: pd.DataFrame = m_info[m_info.fund_id == fund_id]
            rate.loc[:, fund_id] = rate[fund_id].where(partial(ManagerProcessor._lambda_func_1, single_fund_info), 0)
            scale.loc[:, fund_id] = scale[fund_id].where(partial(ManagerProcessor._lambda_func_1, single_fund_info), 0)
        scale = scale + rate - rate  # rate为NaN的地方也让scale变为NaN
        scale = scale.fillna(0)  # 这里在将NaN全变为0
        ratio = (rate * scale).sum(axis=1) / scale.sum(axis=1).replace(0, 1)
        if ratio.shape[0] == 0:
            return
        ratio = pd.Series(np.where(ratio, ratio, index_rate.loc[ratio.index[0]:ratio.index[-1]]), index=ratio.index).fillna(0)
        # 基金经理指数从1开始，即第一日收益为0
        last_trading_day = self.trading_days_list[self.trading_days_list < ratio.index.values[0]].values[-1]
        ratio.loc[last_trading_day] = 0
        ratio = ratio.sort_index()
        return np.exp(np.log(1 + ratio).cumsum())

    def _calc_part3(self, eval_date_dt):
        eval_start_date = (eval_date_dt - datetime.timedelta(days=365))
        self._do_calc_part3(eval_start_date, eval_date_dt)

    def _do_calc_part3(self, eval_start_date, eval_date):
        _rank_list = {}
        _tm_part3_start = time.time()
        fund_size_col = self.fund_size.columns.tolist()
        print(f'eval_start_date {eval_start_date} eval_date {eval_date}')
        for class_name, index_id in self.REPLACE_DICT.items():
            the_eval_start_date = max(eval_start_date, self.DATE_DICT[class_name])
            class_df = self.mng_index_list[class_name]
            bm = self.index_price[index_id]
            fund_ids = self.fund_manager_info['fund_id'].isin(self.class_fund[class_name])
            try:
                cols = class_df.loc[eval_date].dropna().index
            except KeyError:
                continue
            benchmark_stat_result = []
            for i in cols:
                df = class_df[i].loc[the_eval_start_date:eval_date].dropna()
                # 计算年度收益std 如果只有一个， 返回nan
                df_whole = class_df[i].loc[:eval_date].dropna()
                _df = pd.DataFrame(df_whole.rename('mng_id'))
                _df['year'] = _df.index.map(lambda x : x.year)
                year_ret_std = _df.groupby('year').apply( lambda x : (x.iloc[-1]['mng_id'] / x.iloc[0]['mng_id']) - 1).std(ddof=1)
                if df.shape[0] < 5:
                    continue
                dic = Calculator.get_manager_stat_result(df.index.array, df.array, bm.loc[df.index].array)
                dic['name'] = self.manager_name_dict[i]
                dic['index'] = i
                dic['year_ret_std'] = year_ret_std
                fund_df = self.fund_manager_info[self.fund_manager_info['manager_id'].isin([i]) & fund_ids]
                _fund_df = fund_df[fund_df['end'] >= eval_date]
                _fund_list = _fund_df.fund_id.tolist()
                comp_now = _fund_df.company_id.unique()[0]
                _start_com_date = fund_df[fund_df['company_id'] == comp_now].start.min()
                dic['start_date'] = min(_fund_df['start'])
                dic['trade_year'] = (dic['end_date'] - dic['start_date']).days / 365
                _fund_list = [ i for i in _fund_list if i in fund_size_col]
                dic['fund_size'] = self.fund_size.loc[eval_date,_fund_list].sum()
                dic['now_comp_year'] = (dic['end_date'] - _start_com_date).days / 365
                benchmark_stat_result.append(dic)
            stats: pd.DataFrame = pd.DataFrame(benchmark_stat_result).set_index('index')
            rf = 1.025 ** (1 / 365) - 1
            cl_result = []
            for i in stats.index:
                df = class_df[i].loc[the_eval_start_date: eval_date].dropna()
                total = pd.concat([bm.loc[df.index].pct_change(fill_method=None) - rf, df.pct_change(fill_method=None)], axis=1)
                total = total[total.notna().all(axis=1)]
                res = DerivedDataHelper._lambda_cl(total.to_numpy())
                cl_result.append(res)
            temp = pd.DataFrame(cl_result)
            stats = stats.assign(model_alpha=temp['alpha'].array, model_beta=temp['beta'].array)
            # 存进去原始数据而不是rank
            _rank_list[class_name] = stats
        _tm_part3_end = time.time()
        print(f' part3 cost time {_tm_part3_end - _tm_part3_start}')
        res = []
        _tm_part4_start = time.time()
        for class_name in self.REPLACE_DICT.keys():
            stats = _rank_list[class_name].drop(columns=['name', 'start_date', 'end_date'])
            if class_name != 'money':
                stats['score'] = score_rescale(normalize(stats['trade_year']) + normalize(stats['annualized_ret']) + normalize(stats['sharpe']) + normalize(-stats['mdd']) + normalize(stats['model_alpha']) + normalize(stats['model_beta']))
            else:
                stats['score'] = score_rescale(normalize(stats['trade_year']) + normalize(stats['annualized_ret']) + normalize(-stats['annualized_vol']))
            #stats = stats[stats.notna().any(axis=1)]
            stats = stats.rename_axis(index='manager_id').reset_index()
            stats['fund_type'] = class_name
            stats['datetime'] = eval_date
            res.append(stats)
        self._result = pd.concat(res)
        self._data_helper._upload_derived(self._result, FundManagerScore.__table__.name)
        _tm_part4_end = time.time()
        print(f' part4 cost time {_tm_part4_end - _tm_part4_start}')
        print(f'{eval_date} done, cost {_tm_part4_end - _tm_part3_start}')

    def _preprocess_manager_index_df(self, x):
        df = x.drop(columns='fund_type').pivot(index='datetime', columns='manager_id', values='manager_index')
        self.mng_index_list[x.fund_type.array[0]] = df.set_axis(df.index)

    def calc_mng_indicator(self, start_date: str, end_date: str):
        self.init(start_date=start_date, end_date=end_date, is_mng_indicator=True)
        eval_date_dt_list = []
        for eval_date in self.trim_days.sort_values(ascending=False).array:
            eval_date_dt = pd.to_datetime(eval_date, infer_datetime_format=True).date()
            eval_date_dt_list.append(eval_date_dt)
        eval_date_dt_list = [ i for i in eval_date_dt_list if i.weekday() == 4]
        #eval_date_dt_list = [i for i in eval_date_dt_list if i < last_date ]
        #eval_date_dt_list = [i for i in eval_date_dt_list if i > datetime.date(2011,1,1)]
        p = Pool()
        for i in p.imap_unordered(partial(self._calc_part3), eval_date_dt_list, 16):
            pass
        print('all done, to close and join')
        p.close()
        p.join()

    def _do_score_history(self, start_date: str, end_date: str):
        self.trading_days: pd.DataFrame = self._basic_api.get_trading_day_list()
        self.trading_days['datetime'] = self.trading_days.datetime.map(str)
        self.trading_days = self.trading_days[(self.trading_days.datetime >= start_date) & (self.trading_days.datetime <= end_date)]
        print('load td done')

        with DerivedDatabaseConnector().managed_session() as quant_session:
            mng_query = quant_session.query(
                FundManagerInfo.mng_id,
                FundManagerInfo.start_date,
                FundManagerInfo.fund_id,
                FundManagerInfo.mng_name,
                FundManagerInfo.end_date)
            self.fund_manager_info: pd.DataFrame = pd.read_sql(mng_query.statement, mng_query.session.bind)
            self.fund_manager_info = ManagerProcessor._clean_fm(self.fund_manager_info)
            print('load fm done')

        self.index_price: pd.DataFrame = self._basic_api.get_index_price(index_list=list(self.INDEX_LIST.keys()))
        self.index_price = self._clean_index(self.index_price, self.trading_days).loc[start_date:end_date]
        print('load base index done')

        with BasicDatabaseConnector().managed_session() as quant_session:
            fund_info_query = quant_session.query(FundInfo)
        self.fund_info: pd.DataFrame = pd.read_sql(fund_info_query.statement, fund_info_query.session.bind)
        print('load info done')

        fund_manager_index: pd.DataFrame = DerivedDataApi().get_fund_manager_index().drop(columns='_update_time')
        print('load fund manager index done')
        fund_manager_index.groupby(by='fund_type', sort=False).apply(self._preprocess_manager_index_df)
        print('process fund manager index done')

        self.manager_name_dict = dict(zip(self.fund_manager_info['manager_id'].tolist(), self.fund_manager_info['name'].tolist()))
        self.class_fund = {}
        for class_name, index_id in self.REPLACE_DICT.items():
            self.class_fund[class_name] = self.fund_info[self.fund_info['wind_class_2'].isin(FUND_CLASSIFIER[class_name])].fund_id
        print('all data done, to calc fund manager score day by day')

        eval_date_dt_list: List[datetime.date] = []
        min_date = fund_manager_index.datetime.min()
        for eval_date in self.trading_days.datetime.sort_values(ascending=False).array:
            eval_date_dt = pd.to_datetime(eval_date, infer_datetime_format=True).date()
            if eval_date_dt < min_date:
                continue

            eval_date_dt_list.append(eval_date_dt)
            # eval_start_date_dt = eval_date_dt - datetime.timedelta(days=365)
            # self._calc_part3(eval_start_date_dt.isoformat(), eval_date_dt.isoformat(), self.manager_name_dict, self.class_fund)
            # print(f'{eval_date_dt} done')

        p = Pool()
        for i in p.imap_unordered(partial(self._calc_part3, self.manager_name_dict, self.class_fund), eval_date_dt_list, 16):
            pass
        print('all done, to close and join')
        p.close()
        p.join()

    def do_score_update(self, eval_date_dt: datetime.date):
        self.manager_name_dict = dict(zip(self.fund_manager_info['manager_id'].tolist(), self.fund_manager_info['name'].tolist()))
        self.class_fund = {}
        for class_name, index_id in self.REPLACE_DICT.items():
            self.class_fund[class_name] = self.fund_info[self.fund_info['wind_class_2'].isin(FUND_CLASSIFIER[class_name])].fund_id
        self._calc_part3(eval_date_dt)

    def do_index_history(self, start_date: str, end_date: datetime.date, history: bool = False) -> List[str]:
        self.init(start_date, end_date, is_mng_index=True)
        print('init done, to calc manager index')
        self.calc_index()
        for class_type, df in self.mng_index_list.items():
            self.index_to_inserted = df.rename_axis(index='datetime', columns='manager_id').stack().rename('manager_index').reset_index()
            if not history:
                self.index_to_inserted = self.index_to_inserted[self.index_to_inserted.datetime == end_date]
            self.index_to_inserted['fund_type'] = class_type
            self._data_helper._upload_derived(self.index_to_inserted, FundManagerIndex.__table__.name)

    def process(self, end_date: str):
        failed_tasks = []
        try:
            end_date_dt: datetime.date = pd.to_datetime(end_date, infer_datetime_format=True).date()
            self.do_index_history('2005-01-01', end_date_dt)
            self.do_score_update(end_date_dt)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('fund_manager_processor')
        return failed_tasks


if __name__ == '__main__':
    fmis = ManagerProcessor(DerivedDataHelper())
    fmis.process('20210128')
    # fmis.do_index_history(start_date='2005-01-01', end_date='2020-09-07', eval_date='2020-09-07', eval_start_date='2005-01-01')
    # fmis.do_score_history(start_date='2005-01-01', end_date='2020-09-07')



import datetime
import copy
import json
import numpy as np
import pandas as pd
import os
from .order_manager import OrderManager
from .backtest_report import BacktestReport
from ..aws_util import backtest_result_to_dynamodb
from ..data.manager_stock import StockDataManager
from ..match_engine import MatchEngine
from ..structure import AssetType, PositionInfo, BacktestKey
from ..data.manager_stock import StockDailyWrapper
from ..constant import Direction

class BacktestEngine:

    def __init__(self,data_manager=None,data_path=None,start_time=None,end_time=None,universe=None,datatype=None,asset_types=None,stock_pool=None,price_type=None,strategy_type=None,factor_path=None,factor_name=None,vwap_limit=None):
        self._strategies = {}
        self._universe = universe
        self._to_stop = False
        self._current_dt = start_time if isinstance(start_time, datetime.datetime) else datetime.datetime.strptime(start_time,'%Y%m%d')
        self.strategy_type = strategy_type
        
        self._order_manager = OrderManager(self, strategy_type)
        self._match_engine = MatchEngine(self, strategy_type)
        self._asset_types = {}
        self.stock_pool = stock_pool
        if datatype != StockDailyWrapper:
            print(start_time, end_time, universe)
            for ticker in self._universe:
                if isinstance(asset_types, dict):
                    self._asset_types[ticker] = asset_types.get(ticker)
                elif isinstance(asset_types, int):
                    self._asset_types[ticker] = asset_types
                elif asset_types is None:
                    self._asset_types[ticker] = AssetType.STOCK
        else:
            self._all_asset_types = AssetType.STOCK

        self._data_manager = data_manager if data_manager else StockDataManager(data_path    = data_path, 
                                                                                start_time   = start_time, 
                                                                                end_time     = end_time, 
                                                                                universe     = universe,
                                                                                datatype     = datatype,
                                                                                stock_pool   = stock_pool,
                                                                                price_type   = price_type,
                                                                                factor_path  = factor_path,
                                                                                factor_name  = factor_name,
                                                                                vwap_limit   = vwap_limit)

    def _add_strategy(self, strategy, position_info=None):
        strategy.set_context(self)
        self._strategies[strategy._name] = strategy
        self._order_manager.add_strategy(strategy._name, position_info)

    def _init(self):
        self._data_manager.init()
        self._order_manager.init()
        self._match_engine.init()
        for st in self._strategies.values():
            st.init()

    def _terminate(self):
        self._data_manager.terminate()
        self._order_manager.terminate()
        self._match_engine.terminate()

    def _run_daily(self):
        day_data = self._data_manager.get_next_day()
        while day_data is not None:
            self._set_dt(day_data['time'])
            self._match_engine.on_day(day_data)
            self._order_manager.on_day(day_data)
            for st in self._strategies.values():
                st.on_day(day_data)
                # include build order with id, match price, and trade 
                # slippage calculated during match, trade fee calculated in structure.process_trade 
                self._order_manager.process_orders_daily(st.order_list) 

            self._switch_day()                
            day_data = self._data_manager.get_next_day()
                
    def _run_next_tick(self):
        tk = self._data_manager.get_next_tick()
        while tk is not None:
            self._set_dt(self._data_manager.current_index.to_pydatetime())
            self._match_engine.on_tick(tk)
            self._order_manager.on_tick(tk)
            for st in self._strategies.values():
                st.on_tick(tk)
            self._order_manager.process_orders()
            self._order_manager.process_transfers()
            tk = self._data_manager.get_next_tick()

    def _run(self):
        df = self._data_manager.get_data()
        for index, row in df.iterrows():
            if self._to_stop:
                self._terminate()
                return
            cur_dt = pd.to_datetime(index, format='%Y-%m-%dT%H:%M:%S.%fZ')
            self._set_dt(cur_dt)
            tk = self._data_manager.parse_row(index, row)
            self._match_engine.on_tick(tk)
            self._order_manager.on_tick(tk)
            for st in self._strategies.values():
                st.on_tick(tk)
            self._order_manager.process_orders()
            self._order_manager.process_transfers()

    def print_result(self, st_name=None):
        if st_name is None:
            for nm in self._strategies.keys():
                self._order_manager.print_result(nm)
        else:
            self._order_manager.print_result(st_name)      

    def _set_dt(self, dt):
        if self.strategy_type != 'stock_daily':
            if self._current_dt is None or self._current_dt.day != dt.day:
                self._switch_day()
        self._current_dt = dt
    
    def _switch_day(self):
        self._order_manager.switch_day()
        self._match_engine.switch_day()
        for st in self._strategies.values():
            st.switch_day()

    # field member may use
    def _push_trade(self, trade):
        ## called by match_engine
        self._order_manager.process_trade(trade)
        self._strategies[trade.source].on_trade(trade)

    def _push_order_status(self, order):
        self._strategies[order.source].on_order(order.order_id, order.status)
        
    def _push_transfer(self, transfer):
        self._strategies[transfer.source].on_transfer(transfer)

    # utility functions that strategy may use
    def now(self):
        return self._current_dt

    def insert_order(self, order):
        od = copy.deepcopy(order)
        self._order_manager.check_order(od)
        order.order_id = od.order_id
        return None if od.err_msg else od.order_id

    def insert_order_daily(self, order_list):
        self._order_manager.process_orders_daily(order_list)

    def insert_transfer(self, transfer):
        tr = copy.deepcopy(transfer)
        self._order_manager.check_transfer(tr)
        transfer.transfer_id = tr.transfer_id
        return None if tr.err_msg else tr.transfer_id

    def cancel_order(self, order_id):
        res = self._order_manager.cancel_order(order_id)
        return res is True

    def get_position(self, name):
        return self._order_manager.get_position(name)

    def get_trade(self, name):
        return self._order_manager.get_trade(name)

    def get_market_value(self, name):
        return self._order_manager.get_market_value(name)

    def stop(self):
        self._to_stop = True
    
    def save_backtest_result(self):
        self.backtest_result = {}
        for st in self._strategies.values():
            self.backtest_result[st._name] = {}
            trade_data = [[ti.trade_id, ti.ticker, Direction.read(ti.direction), ti.trade_price, ti.trade_volume, ti.time] for ti in self.get_trade(st._name) ]
            self.backtest_result[st._name]['trade'] = {  
                'columns'   :['trade_id','ticker','direction','trade_price','trade_volume','trade_time'],
                'data'      :trade_data,
            }
            self.backtest_result[st._name]['cash'] ={
                'columns'   : ['date', 'cash'],
                'data'      : [[d, int(c)] for d, c in self._order_manager.cash_list[st._name]],
            }
            self.backtest_result[st._name]['pos'] = self._order_manager.pos[st._name]
            self.backtest_result[st._name]['mv'] ={
                'columns'   : ['date', 'mv'],
                'data'      : [[d, int(c)] for d, c in self._order_manager.mvs[st._name]],
            }
            self.backtest_result[st._name]['report'] = BacktestReport( 
                                                            backtest_result = self.backtest_result[st._name],
                                                            period          = self._strategies[st._name].period,
                                                            date_list       = self._data_manager.wrapper.date_list,
                                                            price_data      = self._data_manager.wrapper.price_data,
                                                            factor_data     = self._data_manager.wrapper.factor_data,
                                                            stock_pool_data = self._data_manager.wrapper.stock_pool_data,
                                                            stock_info      = self._data_manager.wrapper.stock_info,
                                                            base_path       = self._data_manager.wrapper.base_path,
                                                            common_pool     = self._data_manager.wrapper.common_pool,
                                                            pool            = self.stock_pool
            ).analysis_backtest_result()

            self.backtest_id = BacktestKey(
                strategy_name   =   st._name,
                factor          =   st.factor_name,
                pool            =   self.stock_pool,
                longshort       =   st.longshort[2],
                top             =   st.top,
                period          =   self._strategies[st._name].period,
                buy_price       =   self._data_manager.wrapper.price_type,
                sell_price      =   self._data_manager.wrapper.price_type,
                fee             =   self.backtest_result[st._name]['report']['total_trade_fee'],
                start           =   self._data_manager.start_time.strftime('%Y-%m_%d'),
                end             =   self._data_manager.end_time.strftime('%Y-%m_%d'),
                vwap_limit      =   self._data_manager.wrapper.vwap_limit,
            ).key()

            self.backtest_result[st._name]['report']['pool']        = self.stock_pool
            self.backtest_result[st._name]['report']['factor_name'] = st.factor_name
            self.backtest_result[st._name]['report']['longshort']   = st.longshort[2]
            self.backtest_result[st._name]['report']['top']         = st.top
            self.backtest_result[st._name]['report']['buy_price']   = self._data_manager.wrapper.price_type
            self.backtest_result[st._name]['report']['sell_price']  = self._data_manager.wrapper.price_type
            self.backtest_result[st._name]['report']['vwap_limit']  = self._data_manager.wrapper.vwap_limit
            self.backtest_result[st._name]['report']['strategy_name'] = st._name
            self.backtest_result[st._name]['report']['backtest_id'] = self.backtest_id

            self.log_path = './backtest_log'
            if not os.path.exists(self.log_path):
                os.makedirs(self.log_path)

            with open(f'{self.log_path}/{st._name}.json','w') as f:
                f.write(json.dumps(self.backtest_result))

            backtest_result_to_dynamodb(self.backtest_result[st._name]['report'])

        
  



import pandas as pd
from datetime import date
from .reader import read_income_quarterly_sheet, read_cash_flow_quarterly_sheet


def convert_quarterly_to_ttm(df: pd.DataFrame):
    """
    根据单季计算TTM
    :param df: columns 必须是日期，index 为财报条目
    :return:
    """
    dfc = df[sorted(list(df.columns))]
    dfc = dfc.rolling(4, min_periods=4, axis=1).sum()

    dfc = dfc.dropna(axis=1, how='all')
    return dfc


def handle_income_ttm(income: pd.DataFrame):
    income = income.T
    income_ttm = convert_quarterly_to_ttm(income)
    income_ttm.index = [x.replace('单季度.', 'TTM.') for x in income_ttm.index]
    return income_ttm.T


def handle_cash_flow_ttm(cash_flow: pd.DataFrame):
    df = cash_flow.T
    df_ttm = convert_quarterly_to_ttm(df)
    df_ttm.index = [x.replace('单季度.', 'TTM.') for x in df_ttm.index]
    return df_ttm.T

